package com.ynet.life.action;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class IdWorker {

   protected static final Logger LOG = LoggerFactory.getLogger(IdWorker.class);
   private long workerId;
   private long datacenterId;
   private long sequence = 0L;
   private long twepoch = 1288834974657L;
   private long workerIdBits = 5L;
   private long datacenterIdBits = 5L;
   private long maxWorkerId;
   private long maxDatacenterId;
   private long sequenceBits;
   private long workerIdShift;
   private long datacenterIdShift;
   private long timestampLeftShift;
   private long sequenceMask;
   private long lastTimestamp;
   private static IdWorker instance = null;


   public static IdWorker getInstance() {
      if(instance == null) {
         instance = new IdWorker(1L, 1L);
      }

      return instance;
   }

   public IdWorker(long workerId, long datacenterId) {
      this.maxWorkerId = ~(-1L << (int)this.workerIdBits);
      this.maxDatacenterId = ~(-1L << (int)this.datacenterIdBits);
      this.sequenceBits = 12L;
      this.workerIdShift = this.sequenceBits;
      this.datacenterIdShift = this.sequenceBits + this.workerIdBits;
      this.timestampLeftShift = this.sequenceBits + this.workerIdBits + this.datacenterIdBits;
      this.sequenceMask = ~(-1L << (int)this.sequenceBits);
      this.lastTimestamp = -1L;
      if(workerId <= this.maxWorkerId && workerId >= 0L) {
         if(datacenterId <= this.maxDatacenterId && datacenterId >= 0L) {
            this.workerId = workerId;
            this.datacenterId = datacenterId;
            LOG.info(String.format("worker starting. timestamp left shift %d, datacenter id bits %d, worker id bits %d, sequence bits %d, workerid %d", new Object[]{Long.valueOf(this.timestampLeftShift), Long.valueOf(this.datacenterIdBits), Long.valueOf(this.workerIdBits), Long.valueOf(this.sequenceBits), Long.valueOf(workerId)}));
         } else {
            throw new IllegalArgumentException(String.format("datacenter Id can\'t be greater than %d or less than 0", new Object[]{Long.valueOf(this.maxDatacenterId)}));
         }
      } else {
         throw new IllegalArgumentException(String.format("worker Id can\'t be greater than %d or less than 0", new Object[]{Long.valueOf(this.maxWorkerId)}));
      }
   }

   public synchronized long nextId() {
      long timestamp;
      if((timestamp = this.timeGen()) < this.lastTimestamp) {
         LOG.error(String.format("clock is moving backwards.  Rejecting requests until %d.", new Object[]{Long.valueOf(this.lastTimestamp)}));
         throw new RuntimeException(String.format("Clock moved backwards.  Refusing to generate id for %d milliseconds", new Object[]{Long.valueOf(this.lastTimestamp - timestamp)}));
      } else {
         if(this.lastTimestamp == timestamp) {
            this.sequence = this.sequence + 1L & this.sequenceMask;
            if(this.sequence == 0L) {
               timestamp = this.tilNextMillis(this.lastTimestamp);
            }
         } else {
            this.sequence = 0L;
         }

         this.lastTimestamp = timestamp;
         return timestamp - this.twepoch << (int)this.timestampLeftShift | this.datacenterId << (int)this.datacenterIdShift | this.workerId << (int)this.workerIdShift | this.sequence;
      }
   }

   protected long tilNextMillis(long lastTimestamp) {
      long timestamp;
      while((timestamp = this.timeGen()) <= lastTimestamp) {
         ;
      }

      return timestamp;
   }

   protected long timeGen() {
      return System.currentTimeMillis();
   }

   public static void main(String[] args) {
      System.out.println(getInstance().nextId());
   }
}
